/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil -*- */
/*
 * Copyright (c) 2013-2014 The University of Tennessee and The University
 *                         of Tennessee Research Foundation.  All rights
 *                         reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <math.h>
#include <mpi.h>

void print_iterations( MPI_Comm scomm, int i );
int rank, verbose=0; /* makes this global (for printfs) */
#define COUNT 1024

int main( int argc, char* argv[] ) { 
    MPI_Comm fcomm, scomm; /* a comm to inject a failure, and a safe comm */
    int np, victim; /* the victim rank */
    int rc; /* error code from MPI functions */
    char estr[MPI_MAX_ERROR_STRING]=""; int strl; /* error messages */
    double start, tff, twf; /* timings */
    double sarray[COUNT], rarray[COUNT];
    int i;
    
    MPI_Init( &argc, &argv );
    MPI_Comm_size( MPI_COMM_WORLD, &np );
    MPI_Comm_rank( MPI_COMM_WORLD, &rank );
    if( !strcmp( argv[argc-1], "-v" ) ) verbose=1;

    /* The victim is always the last process (for simplicity) */
    victim = (rank == np-1)? 1 : 0;

    /* Now, we need a communicator that still works after the failure:
     *  this split creates a communicator that excludes the victim;
     *  we can do this, because we know the victim a-priori, in this
     *  example. */
    MPI_Comm_split( MPI_COMM_WORLD, victim, rank, &scomm );

    /* Let's work on a copy of MPI_COMM_WORLD, good practice anyway. */
    MPI_Comm_dup( MPI_COMM_WORLD, &fcomm );

    /* We set an errhandler on fcomm, so that a failure is not fatal anymore. 
     * Note: the errhandler on scomm and MPI_COMM_WORLD are still the
     *   default: if an operation involves a failed process on these
     *   communicators, the program still aborts! */
    MPI_Comm_set_errhandler( fcomm, MPI_ERRORS_RETURN );

    for( i = 0; i < 10; i++ ) {
        /* Victim suicides */
        if( 3 == i && victim ) {
            printf( "Rank %04d: committing suicide at iteration %d\n", rank, i );
            raise( SIGKILL );
            while(1); /* wait for the signal */
        }

        if(verbose) printf( "Rank %04d: entering Sendrecv %d\n", rank, i );
        start=MPI_Wtime();
    	/* Point to point replaced collective operations compared to the previous 
		 * program. */
	    rc = MPI_Sendrecv( sarray, COUNT, MPI_DOUBLE, (rank+1)%np, 0,
                           rarray, COUNT, MPI_DOUBLE, (rank-1)%np, 0,
                           fcomm, MPI_STATUS_IGNORE );
        twf=MPI_Wtime()-start;
        MPI_Error_string( rc, estr, &strl );
        printf( "Rank %04d: Sendrecv %d completed (rc=%s) duration %g (s)\n", rank, i, estr, twf );

        if( rc != MPI_SUCCESS ) { 
            /* ???>>> Hu ho, this program has a problem here, can you tell? ???<< */
		    goto cleanup;
        }
    }

cleanup:
    print_iterations( scomm, i-1 );

    MPI_Comm_free( &fcomm );
    MPI_Comm_free( &scomm );

    MPI_Finalize();
    return EXIT_SUCCESS;
}

void print_iterations( MPI_Comm scomm, 
                       int i ) {
    /* Storage for min and max times */
    int mi, Mi;

    MPI_Reduce( &i, &mi, 1, MPI_INTEGER, MPI_MIN, 0, scomm );
    MPI_Reduce( &i, &Mi, 1, MPI_INTEGER, MPI_MAX, 0, scomm );

	/* >>??? What do you expect about iterations here? ???<< */
    if( 0 == rank ) printf( 
        "## Iterations ### Min         ### Max         ##\n"
        "Sendrecv       # %13d # %13d \n"
        , mi, Mi );
}

